from molang import compile


def test_alias():
    a = compile("c.test")
    b = compile("q.test")
    c = compile("t.test")
    d = compile("v.test")

    e = compile("context.test")
    f = compile("query.test")
    g = compile("temp.test")
    h = compile("variable.test")

    assert a == "context.test"
    assert b == "query.test"
    assert c == "temp.test"
    assert d == "variable.test"

    assert e == "context.test"
    assert f == "query.test"
    assert g == "temp.test"
    assert h == "variable.test"



from argparse import ArgumentParser
from molang import __version__

parser = ArgumentParser(prog="molang", description="Run molang files")
# parser.add_argument('file', metavar='<file>', nargs='?', const=None, help='execute the molang code contained in a file.')
# parser.add_argument('-c', type=str, metavar='<cmd>' , dest='cmd', help='execute the molang code in cmd.
#   cmd can be one or more statements separated by newlines, with significant leading whitespace as in normal module code.')
parser.add_argument(
    "-V",
    "--version",
    action="store_true",
    help="print the molang version number and exit.",
)


def main():
    args = parser.parse_args()
    if args.version:
        print(__version__)
    # elif args.file:
    #     run_file(args.file)
    # elif args.cmd:
    #     run('<stdin>', args.cmd)


if __name__ == "__main__":
    main()



__all__ = [
    "MolangExpr",
    "array",
    "geometry",
    "material",
    "math",
    "query",
    "temp",
    "texture",
    "variable",
    "context",
]

from typing import Any, Dict
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from .core.ast import Node
import molang


class MolangExpr:
    def __init__(self, expr: Any):
        self.expr = MolangExpr.convert(expr)

    def eval(self, context: Dict[str, Any]) -> Any:
        return molang.eval(self.expr, context)

    def parse(self) -> Node:
        return molang.parse(self.expr)

    def compile(self) -> str:
        return molang.compile(self.expr)

    @staticmethod
    def convert(value: str | bool | float | int) -> str:
        if isinstance(value, bool):
            return "1" if value else "0"
        text = str(value)
        match text.lower():
            case "true":
                return "1"
            case "false":
                return "0"
        return text

    @staticmethod
    def value(value: Any) -> "str|MolangExpr":
        if isinstance(value, MolangExpr):
            return value
        x = repr(value)
        return MolangExpr.convert(x)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls,
            core_schema.union_schema(
                [
                    core_schema.is_instance_schema(cls),
                    core_schema.chain_schema(
                        [
                            core_schema.str_schema(),
                            core_schema.no_info_plain_validator_function(
                                lambda v: cls(v)
                            ),
                        ]
                    ),
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: str(instance),
                return_schema=core_schema.str_schema(),
            ),
        )

    def _bin(self, op: str, other: Any) -> "MolangExpr":
        return MolangExpr(f"({self} {op} {MolangExpr.value(other)})")

    def _log(self, op: str, other: Any) -> "MolangExpr":
        return MolangExpr(f"{self} {op} {MolangExpr.value(other)}")

    def __str__(self) -> str:
        return self.expr

    def __repr__(self) -> str:
        return self.expr

    def __call__(self, *args) -> "MolangExpr":
        return MolangExpr(
            f"{self}({', '.join([str(MolangExpr.value(x)) for x in args])})"
        )

    def __invert__(self) -> "MolangExpr":
        return MolangExpr(f"!{self}")

    def __or__(self, other: Any) -> "MolangExpr":
        return self._log("||", other)

    def __and__(self, other: Any) -> "MolangExpr":
        return self._log("&&", other)

    def __gt__(self, other: Any) -> "MolangExpr":
        return self._log(">", other)

    def __lt__(self, other: Any) -> "MolangExpr":
        return self._log("<", other)

    def __ge__(self, other: Any) -> "MolangExpr":
        return self._log(">=", other)

    def __le__(self, other: Any) -> "MolangExpr":
        return self._log("<=", other)

    def __ne__(self, other: Any) -> "MolangExpr":  # type: ignore
        return self._log("!=", other)

    def __eq__(self, other: Any) -> "MolangExpr":  # type: ignore
        return self._log("==", other)

    def __add__(self, other: Any) -> "MolangExpr":
        return self._bin("+", other)

    def __sub__(self, other: Any) -> "MolangExpr":
        return self._bin("-", other)

    def __mul__(self, other: Any) -> "MolangExpr":
        return self._bin("*", other)

    def __truediv__(self, other: Any) -> "MolangExpr":
        return self._bin("/", other)

    def __floordiv__(self, other: Any) -> "MolangExpr":
        return self._bin("/", other)

    def __getitem__(self, name: "float|int|MolangExpr") -> "MolangExpr":
        return MolangExpr(f"{self}[{name}]")

    def if_(self, true_v: Any, false_v: Any) -> "MolangExpr":
        return MolangExpr(
            f"({self} ? {MolangExpr.value(true_v)} : {MolangExpr.value(false_v)})"
        )

    @staticmethod
    def for_each(
        variable: "MolangExpr", array: "MolangExpr", expression: "MolangExpr"
    ) -> "MolangExpr":
        return MolangExpr(f"for_each({variable}, {array}, {{{expression}}});")

    @staticmethod
    def loop(count: "int|MolangExpr", expression: "MolangExpr") -> "MolangExpr":
        return MolangExpr(f"loop({count}, {{{expression}}});")


class ArrayNamespace:
    def __getattr__(self, name: str) -> MolangExpr:
        return MolangExpr(f"array.{name.lower()}")


class VariableNamespace:
    _assignments: Dict[str, Any] = {}

    def __call__(self, name: str, *args) -> MolangExpr:
        if len(args) == 0:
            return MolangExpr(f"{self._prefix}.{name.lower()}")
        return MolangExpr(f"{self._prefix}.{name.lower()}")(*args)

    def __init__(self, prefix: str):
        self._prefix = prefix

    def __getattr__(self, name: str) -> MolangExpr:
        name = name.lower()
        if name in self._assignments:
            v = self._assignments[name]
            return MolangExpr(f"{self._prefix}.{name} = {MolangExpr.value(v)};")
        return MolangExpr(f"{self._prefix}.{name}")

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            object.__setattr__(self, name, value)
            return
        self._assignments[name.lower()] = value


class MathNamespace(VariableNamespace):
    pass


class QueryNamespace(VariableNamespace):
    pass


variable = VariableNamespace("variable")
temp = VariableNamespace("temp")
context = VariableNamespace("context")
geometry = VariableNamespace("geometry")
texture = VariableNamespace("texture")
material = VariableNamespace("material")
query = QueryNamespace("query")
math = MathNamespace("math")
array = ArrayNamespace()



from molang import eval, create_context


def test_eval_query():
    a = eval("query.health")
    b = eval("query.health", {"query": {"health": 100}})
    assert a == 0
    assert b == 100


def test_eval_integer():
    a = eval("100")
    b = eval("-100")
    assert a == 100
    assert b == -100


def test_eval_float():
    a = eval("3.14159")
    b = eval("-3.14159")
    assert a == 3.14159
    assert b == -3.14159


def test_eval_assign():
    ctx = create_context()
    eval("variable.x = 5", ctx)
    assert ctx["variable"]["x"] == 5


# def test_eval_array():
#     ctx = {'array': {'my_geometries': ["item1", "item2", "item3"]}}
#     a = eval(array.my_geometries[query.anim_time], ctx)
#     b = eval(array.my_geometries[0], ctx)
#     assert a == 'item1'
#     assert b == 'item1'


def test_eval_not():
    """not"""
    a = eval("!query.health")
    assert a == 1


def test_eval_and():
    """and"""
    a = eval("0 && 0")
    b = eval("1 && 1")
    # c = eval('1 && 0')
    d = eval("0 && 1")
    assert not a
    assert b
    # assert not c
    assert not d


def test_eval_or():
    """or"""
    a = eval("0 || 0")
    b = eval("1 || 1")
    # c = eval('1 || 0')
    d = eval("0 || 1")
    assert not a
    assert b
    # assert c
    assert d


def test_eval_lt():
    """less than"""
    a = eval("0 < 100")
    b = eval("100 < 0")
    c = eval("100 < 100")
    assert a
    assert not b
    assert not c


def test_eval_le():
    """less than or equal to"""
    a = eval("0 <= 100")
    b = eval("100 <= 0")
    c = eval("100 <= 100")
    assert a
    assert not b
    assert c


def test_eval_gt():
    """greater than"""
    a = eval("0 > 100")
    # b = eval('100 > 0')
    c = eval("100 > 100")
    assert not a
    # assert b
    assert not c


def test_eval_ge():
    """greater than or equal to"""
    a = eval("0 >= 100")
    # b = eval('100 >= 0')
    c = eval("100 >= 100")
    assert not a
    # assert b
    assert c


def test_eval_eq():
    """equal"""
    a = eval("100 == 100")
    b = eval("0 == 100")
    assert a
    assert not b


def test_eval_ne():
    """not equal"""
    a = eval("100 != 100")
    b = eval("0 != 100")
    assert not a
    assert b


def test_eval_mul():
    a = eval("5 * 2")
    b = eval("5 * 0.10")
    assert a == 10
    assert b == 0.5


def test_eval_truediv():
    a = eval("10 / 2")
    b = eval("10 / 0.10")
    assert a == 5
    assert b == 100


def test_eval_add():
    a = eval("5 + 5")
    b = eval("5 + 0.10")
    assert a == 10
    assert b == 5.10


def test_eval_sub():
    a = eval("5 - 5")
    b = eval("5 - 10")
    c = eval("5 - 0.10")
    d = eval("5 - 10.10")
    assert a == 0
    assert b == -5
    assert c == 4.9
    assert d == -5.1


def test_eval_math():
    """Test all math functions."""
    import math as py_math

    # Constants
    x = eval("math.pi")
    assert x == py_math.pi

    # Basic functions
    assert eval("math.abs(-5)") == 5
    assert eval("math.abs(5)") == 5
    assert eval("math.sqrt(16)") == 4.0
    assert eval("math.pow(2, 3)") == 8
    assert abs(eval("math.exp(1)") - py_math.e) < 1e-9
    assert eval("math.ln(1)") == 0.0

    # Trigonometric
    assert abs(eval("math.sin(0)") - 0) < 1e-9
    assert abs(eval("math.cos(0)") - 1) < 1e-9
    assert abs(eval("math.atan(0)") - 0) < 1e-9
    assert abs(eval("math.asin(0)") - 0) < 1e-9
    assert abs(eval("math.acos(1)") - 0) < 1e-9

    # Atan2
    assert abs(eval("math.atan2(1, 1)") - py_math.pi / 4) < 1e-9

    # Rounding
    assert eval("math.ceil(3.2)") == 4
    assert eval("math.floor(3.8)") == 3
    assert eval("math.round(3.5)") == 4
    assert eval("math.trunc(3.9)") == 3

    # Utility
    assert eval("math.max(3, 7)") == 7
    assert eval("math.min(3, 7)") == 3
    assert eval("math.clamp(5, 0, 10)") == 5
    assert eval("math.clamp(-5, 0, 10)") == 0
    assert eval("math.clamp(15, 0, 10)") == 10

    # Sign
    assert eval("math.sign(5)") == 1
    assert eval("math.sign(-5)") == -1
    assert eval("math.sign(0)") == 0

    # Copy sign
    assert eval("math.copy_sign(5, -1)") == -5
    assert eval("math.copy_sign(-5, 1)") == 5

    # Modulo
    assert eval("math.mod(10, 3)") == 1
    assert eval("math.mod(7, 2)") == 1

    # Min angle
    result = eval("math.min_angle(3.14159)")
    assert -py_math.pi <= result <= py_math.pi

    # Interpolation
    assert abs(eval("math.lerp(0, 10, 0.5)") - 5.0) < 1e-9
    assert abs(eval("math.inverse_lerp(0, 10, 5)") - 0.5) < 1e-9
    assert abs(eval("math.lerprotate(0, 1, 0.5)") - 0.5) < 1e-9
    assert abs(eval("math.hermite_blend(0.5)") - 0.5) < 1e-9

    # Easing functions - test endpoints
    assert abs(eval("math.ease_in_quad(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_quad(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_quad(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_quad(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_quad(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_quad(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_cubic(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_cubic(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_cubic(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_cubic(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_cubic(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_cubic(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_quart(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_quart(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_quart(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_quart(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_quart(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_quart(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_quint(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_quint(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_quint(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_quint(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_quint(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_quint(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_sine(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_sine(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_sine(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_sine(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_sine(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_sine(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_circ(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_circ(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_circ(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_circ(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_circ(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_circ(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_expo(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_expo(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_expo(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_expo(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_expo(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_expo(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_back(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_back(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_back(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_back(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_back(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_back(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_bounce(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_bounce(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_bounce(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_bounce(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_bounce(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_bounce(0, 10, 1)") - 10) < 1e-6

    assert abs(eval("math.ease_in_elastic(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_elastic(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_out_elastic(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_out_elastic(0, 10, 1)") - 10) < 1e-6
    assert abs(eval("math.ease_in_out_elastic(0, 10, 0)") - 0) < 1e-6
    assert abs(eval("math.ease_in_out_elastic(0, 10, 1)") - 10) < 1e-6

    # Random functions - just test that they're callable and return numbers
    result = eval("math.random(0, 10)")
    assert 0 <= result <= 10

    result = eval("math.random_integer(1, 6)")
    assert 1 <= result <= 6

    # Die roll returns average
    result = eval("math.die_roll(1, 1, 6)")
    assert 1 <= result <= 6

    result = eval("math.die_roll_integer(1, 1, 6)")
    assert 1 <= result <= 6


def test_eval_ternary():
    text = "q.is_on_ground ? 12 : 32"
    a = eval(text, {"query": {"is_on_ground": True}})
    b = eval(text, {"query": {"is_on_ground": False}})
    assert a == 12
    assert b == 32



from molang.dsl import (
    MolangExpr,
    array,
    math,
    query,
    variable,
    geometry,
    temp,
    context,
    texture,
    material,
)
from pydantic import BaseModel


def test_query():
    a = query.health
    b = query.HEALTH
    c = query.block_state("facing")
    assert a.expr == "query.health"
    assert b.expr == "query.health"
    assert c.expr == "query.block_state('facing')"

    # Custom

    d = query.custom
    e = query.custom_call("arg1")
    assert d.expr == "query.custom"
    assert e.expr == "query.custom_call('arg1')"


def test_variable():
    a = variable.test
    b = temp.test
    c = context.test
    d = geometry.test
    e = texture.test
    f = material.test
    assert a.expr == "variable.test"
    assert b.expr == "temp.test"
    assert c.expr == "context.test"
    assert d.expr == "geometry.test"
    assert e.expr == "texture.test"
    assert f.expr == "material.test"


def test_variable_call():
    a = variable.test("arg1", 1, 0.5, True)
    b = temp.test("arg1", 1, 0.5, True)
    c = context.test("arg1", 1, 0.5, True)
    d = geometry.test("arg1", 1, 0.5, True)
    e = texture.test("arg1", 1, 0.5, True)
    f = material.test("arg1", 1, 0.5, True)
    assert a.expr == "variable.test('arg1', 1, 0.5, 1)"
    assert b.expr == "temp.test('arg1', 1, 0.5, 1)"
    assert c.expr == "context.test('arg1', 1, 0.5, 1)"
    assert d.expr == "geometry.test('arg1', 1, 0.5, 1)"
    assert e.expr == "texture.test('arg1', 1, 0.5, 1)"
    assert f.expr == "material.test('arg1', 1, 0.5, 1)"


def test_type_integer():
    a = MolangExpr(100)
    b = MolangExpr(-100)
    assert a.expr == "100"
    assert b.expr == "-100"


def test_type_float():
    a = MolangExpr(3.14159)
    b = MolangExpr(-3.14159)
    assert a.expr == "3.14159"
    assert b.expr == "-3.14159"


def test_type_string():
    a = MolangExpr("'hello'")
    b = MolangExpr("'HELLO'")
    assert a.expr == "'hello'"
    assert b.expr == "'HELLO'"


def test_type_boolean():
    a = MolangExpr("true")
    b = MolangExpr("false")
    c = MolangExpr(True)
    d = MolangExpr(False)
    assert a.expr == "1"
    assert b.expr == "0"
    assert c.expr == "1"
    assert d.expr == "0"


def test_type_array():
    a = array.my_geometries[query.anim_time]
    b = array.my_geometries[0]
    c = array.my_geometries[0.0]
    assert a.expr == "array.my_geometries[query.anim_time]"
    assert b.expr == "array.my_geometries[0]"
    assert c.expr == "array.my_geometries[0.0]"


def test_logical_not():
    """not"""
    a = ~query.health
    assert a.expr == "!query.health"


def test_logical_and():
    """and"""
    a = query.health & 100
    b = query.health & query.is_on_ground
    assert a.expr == "query.health && 100"
    assert b.expr == "query.health && query.is_on_ground"


def test_logical_or():
    """or"""
    a = query.health | 100
    b = query.health | query.is_on_ground
    assert a.expr == "query.health || 100"
    assert b.expr == "query.health || query.is_on_ground"


def test_logical_lt():
    """less than"""
    a = query.health < 100
    b = query.health < query.is_on_ground
    assert a.expr == "query.health < 100"
    assert b.expr == "query.health < query.is_on_ground"


def test_logical_le():
    """less than or equal to"""
    a = query.health <= 100
    b = query.health <= query.is_on_ground
    assert a.expr == "query.health <= 100"
    assert b.expr == "query.health <= query.is_on_ground"


def test_logical_gt():
    """greater than"""
    a = query.health > 100
    b = query.health > query.is_on_ground
    assert a.expr == "query.health > 100"
    assert b.expr == "query.health > query.is_on_ground"


def test_logical_ge():
    """greater than or equal to"""
    a = query.health >= 100
    b = query.health >= query.is_on_ground
    assert a.expr == "query.health >= 100"
    assert b.expr == "query.health >= query.is_on_ground"


def test_logical_eq():
    """equal"""
    a = query.health == 100
    b = query.health == query.is_on_ground
    assert a.expr == "query.health == 100"
    assert b.expr == "query.health == query.is_on_ground"


def test_logical_ne():
    """not equal"""
    a = query.health != 100
    b = query.health != query.is_on_ground
    assert a.expr == "query.health != 100"
    assert b.expr == "query.health != query.is_on_ground"


def test_bin_mul():
    a = query.health * 100
    b = query.health * query.is_on_ground
    assert a.expr == "(query.health * 100)"
    assert b.expr == "(query.health * query.is_on_ground)"


def test_bin_truediv():
    a = query.health / 100
    b = query.health / query.is_on_ground
    assert a.expr == "(query.health / 100)"
    assert b.expr == "(query.health / query.is_on_ground)"


def test_bin_floordiv():
    a = query.health // 100
    b = query.health // query.is_on_ground
    assert a.expr == "(query.health / 100)"
    assert b.expr == "(query.health / query.is_on_ground)"


def test_bin_add():
    a = query.health + 100
    b = query.health + query.is_on_ground
    assert a.expr == "(query.health + 100)"
    assert b.expr == "(query.health + query.is_on_ground)"


def test_bin_sub():
    a = query.health - 100
    b = query.health - query.is_on_ground
    assert a.expr == "(query.health - 100)"
    assert b.expr == "(query.health - query.is_on_ground)"


def test_math():
    x = math.pi
    assert x.expr == "math.pi"

    x = math.abs(1)
    assert x.expr == "math.abs(1)"

    x = math.acos(1)
    assert x.expr == "math.acos(1)"

    x = math.asin(1)
    assert x.expr == "math.asin(1)"

    x = math.atan(1)
    assert x.expr == "math.atan(1)"

    x = math.atan2(1, 2)
    assert x.expr == "math.atan2(1, 2)"

    x = math.ceil(1)
    assert x.expr == "math.ceil(1)"

    x = math.clamp(1, 2, 3)
    assert x.expr == "math.clamp(1, 2, 3)"

    x = math.copy_sign(1, 2)
    assert x.expr == "math.copy_sign(1, 2)"

    x = math.cos(1)
    assert x.expr == "math.cos(1)"

    x = math.die_roll(1, 2, 3)
    assert x.expr == "math.die_roll(1, 2, 3)"

    x = math.die_roll_integer(1, 2, 3)
    assert x.expr == "math.die_roll_integer(1, 2, 3)"

    x = math.ease_in_back(1, 2, 3)
    assert x.expr == "math.ease_in_back(1, 2, 3)"

    x = math.ease_in_bounce(1, 2, 3)
    assert x.expr == "math.ease_in_bounce(1, 2, 3)"

    x = math.ease_in_circ(1, 2, 3)
    assert x.expr == "math.ease_in_circ(1, 2, 3)"

    x = math.ease_in_cubic(1, 2, 3)
    assert x.expr == "math.ease_in_cubic(1, 2, 3)"

    x = math.ease_in_elastic(1, 2, 3)
    assert x.expr == "math.ease_in_elastic(1, 2, 3)"

    x = math.ease_in_expo(1, 2, 3)
    assert x.expr == "math.ease_in_expo(1, 2, 3)"

    x = math.ease_in_out_back(1, 2, 3)
    assert x.expr == "math.ease_in_out_back(1, 2, 3)"

    x = math.ease_in_out_bounce(1, 2, 3)
    assert x.expr == "math.ease_in_out_bounce(1, 2, 3)"

    x = math.ease_in_out_circ(1, 2, 3)
    assert x.expr == "math.ease_in_out_circ(1, 2, 3)"

    x = math.ease_in_out_cubic(1, 2, 3)
    assert x.expr == "math.ease_in_out_cubic(1, 2, 3)"

    x = math.ease_in_out_elastic(1, 2, 3)
    assert x.expr == "math.ease_in_out_elastic(1, 2, 3)"

    x = math.ease_in_out_expo(1, 2, 3)
    assert x.expr == "math.ease_in_out_expo(1, 2, 3)"

    x = math.ease_in_out_quad(1, 2, 3)
    assert x.expr == "math.ease_in_out_quad(1, 2, 3)"

    x = math.ease_in_out_quart(1, 2, 3)
    assert x.expr == "math.ease_in_out_quart(1, 2, 3)"

    x = math.ease_in_out_quint(1, 2, 3)
    assert x.expr == "math.ease_in_out_quint(1, 2, 3)"

    x = math.ease_in_out_sine(1, 2, 3)
    assert x.expr == "math.ease_in_out_sine(1, 2, 3)"

    x = math.ease_in_quad(1, 2, 3)
    assert x.expr == "math.ease_in_quad(1, 2, 3)"

    x = math.ease_in_quart(1, 2, 3)
    assert x.expr == "math.ease_in_quart(1, 2, 3)"

    x = math.ease_in_quint(1, 2, 3)
    assert x.expr == "math.ease_in_quint(1, 2, 3)"

    x = math.ease_in_sine(1, 2, 3)
    assert x.expr == "math.ease_in_sine(1, 2, 3)"

    x = math.ease_out_back(1, 2, 3)
    assert x.expr == "math.ease_out_back(1, 2, 3)"

    x = math.ease_out_bounce(1, 2, 3)
    assert x.expr == "math.ease_out_bounce(1, 2, 3)"

    x = math.ease_out_circ(1, 2, 3)
    assert x.expr == "math.ease_out_circ(1, 2, 3)"

    x = math.ease_out_cubic(1, 2, 3)
    assert x.expr == "math.ease_out_cubic(1, 2, 3)"

    x = math.ease_out_elastic(1, 2, 3)
    assert x.expr == "math.ease_out_elastic(1, 2, 3)"

    x = math.ease_out_expo(1, 2, 3)
    assert x.expr == "math.ease_out_expo(1, 2, 3)"

    x = math.ease_out_quad(1, 2, 3)
    assert x.expr == "math.ease_out_quad(1, 2, 3)"

    x = math.ease_out_quart(1, 2, 3)
    assert x.expr == "math.ease_out_quart(1, 2, 3)"

    x = math.ease_out_quint(1, 2, 3)
    assert x.expr == "math.ease_out_quint(1, 2, 3)"

    x = math.ease_out_sine(1, 2, 3)
    assert x.expr == "math.ease_out_sine(1, 2, 3)"

    x = math.exp(1)
    assert x.expr == "math.exp(1)"

    x = math.floor(1)
    assert x.expr == "math.floor(1)"

    x = math.hermite_blend(1)
    assert x.expr == "math.hermite_blend(1)"

    x = math.inverse_lerp(1, 2, 3)
    assert x.expr == "math.inverse_lerp(1, 2, 3)"

    x = math.lerp(1, 2, 3)
    assert x.expr == "math.lerp(1, 2, 3)"

    x = math.lerprotate(1, 2, 3)
    assert x.expr == "math.lerprotate(1, 2, 3)"

    x = math.ln(1)
    assert x.expr == "math.ln(1)"

    x = math.max(1, 2)
    assert x.expr == "math.max(1, 2)"

    x = math.min(1, 2)
    assert x.expr == "math.min(1, 2)"

    x = math.min_angle(1)
    assert x.expr == "math.min_angle(1)"

    x = math.mod(1, 2)
    assert x.expr == "math.mod(1, 2)"

    x = math.pow(1, 2)
    assert x.expr == "math.pow(1, 2)"

    x = math.random(1, 2)
    assert x.expr == "math.random(1, 2)"

    x = math.random_integer(1, 2)
    assert x.expr == "math.random_integer(1, 2)"

    x = math.round(1)
    assert x.expr == "math.round(1)"

    x = math.sign(1)
    assert x.expr == "math.sign(1)"

    x = math.sin(1)
    assert x.expr == "math.sin(1)"

    x = math.sqrt(1)
    assert x.expr == "math.sqrt(1)"

    x = math.trunc(1)
    assert x.expr == "math.trunc(1)"

    # Custom

    x = math.custom(1)
    assert x.expr == "math.custom(1)"


def test_ternary():
    x = query.is_on_ground.if_(12, 32)
    assert x.expr == "(query.is_on_ground ? 12 : 32)"


def test_assign():
    variable.test = 4
    assert variable.test.expr == "variable.test = 4;"

    variable.test = "Hello, World!"
    assert variable.test.expr == "variable.test = 'Hello, World!';"


def test_itr():
    x = MolangExpr.loop(10, MolangExpr("t.x = v.x + v.y;v.x = v.y; v.y = t.x;"))
    assert x.expr == "loop(10, {t.x = v.x + v.y;v.x = v.y; v.y = t.x;});"

    x = MolangExpr.for_each(
        temp.pig,
        query.get_nearby_entities(4, "minecraft:pig"),
        MolangExpr("variable.x = variable.x + 1;"),
    )
    assert (
        x.expr
        == "for_each(temp.pig, query.get_nearby_entities(4, 'minecraft:pig'), {variable.x = variable.x + 1;});"
    )


def test_expr():
    x = ~query.equipped_item_all_tags("slot.armor.chest", "paper") | query.is_on_ground
    assert (
        x.expr
        == "!query.equipped_item_all_tags('slot.armor.chest', 'paper') || query.is_on_ground"
    )

    x = query.block_state("minecraft:cardinal_direction") == "north"
    assert x.expr == "query.block_state('minecraft:cardinal_direction') == 'north'"

    x = (
        math.cos(query.anim_time * 38) * variable.rotation_scale
        + variable.x * variable.x * query.life_time
    )
    assert (
        x.expr
        == "((math.cos((query.anim_time * 38)) * variable.rotation_scale) + ((variable.x * variable.x) * query.life_time))"
    )

    x = query.is_sleeping.if_(
        geometry.my_sleeping_geo,
        array.my_geos[math.cos(query.anim_time * 12.3 + 41.9) * 10 + 0.6],
    )
    assert (
        x.expr
        == "(query.is_sleeping ? geometry.my_sleeping_geo : array.my_geos[((math.cos(((query.anim_time * 12.3) + 41.9)) * 10) + 0.6)])"
    )

    # v.x = 0;
    # for_each(v.pig, query.get_nearby_entities(4, 'minecraft:pig'), {
    #     v.x = v.x + v.pig->query.get_relative_block_state(0, 1, 0, 'flammable');
    # });
    #
    # (v.moo > 0) ? {
    #   v.x = math.sin(q.life_time * 45);
    #   v.x = v.x * v.x + 17.3;
    #   t.sin_x = math.sin(v.x);
    #   v.x = t.sin_x * t.sin_x + v.x * v.x;
    #   v.x = math.sqrt(v.x) * v.x * math.pi;
    # }
    #
    # v.x = 1;
    # v.y = 1;
    # loop(10, {
    #   t.x = v.x + v.y;
    #   v.x = v.y;
    #   v.y = t.x;
    # });
    #
    # v.x = 0;
    # for_each(t.pig, query.get_nearby_entities(4, 'minecraft:pig'), {
    #     v.x = v.x + 1;
    # });
    #
    # v.x = 1;
    # v.y = 1;
    # loop(10, {t.x = v.x + v.y; v.x = v.y; v.y = t.x; (v.y > 20) ? break;});
    #
    # v.x = 0;
    # loop(10, {
    #   (v.x > 5) ? continue;
    #   v.x = v.x + 1;
    # });
    #
    # x = variable.x = (variable.x or 1.2) + 0.3
    #
    # index = max(0, expression_result) % array_size

    x = array.hair_colors[variable.hair_color]
    assert x.expr == "array.hair_colors[variable.hair_color]"

    x = query.is_baby.if_(-8.0, 0.0)
    assert x.expr == "(query.is_baby ? -8.0 : 0.0)"

    x = math.sin(query.anim_time * 1.23)
    print(x)
    assert x.expr == "math.sin((query.anim_time * 1.23))"


def test_pydantic():
    class Test(BaseModel):
        condition: MolangExpr

    x = Test(condition=query.block_state("facing") == "north")
    assert x.model_dump() == {"condition": "query.block_state('facing') == 'north'"}



"""
Domain Specific Language
"""

from .core.ast import Node
from typing import Any

__all__ = [
    "MolangExpr",
    "array",
    "geometry",
    "material",
    "math",
    "query",
    "temp",
    "texture",
    "variable",
    "context",
]

class MolangExpr:
    expr: str
    def __init__(self, expr: Any) -> None: ...
    def eval(self, context: dict[str, Any]) -> Any: ...
    def parse(self) -> Node: ...
    def compile(self) -> str: ...
    @staticmethod
    def value(value: Any) -> str | MolangExpr: ...
    def __call__(self, *args) -> MolangExpr: ...
    def __invert__(self) -> MolangExpr: ...
    def __or__(self, other: Any) -> MolangExpr: ...
    def __and__(self, other: Any) -> MolangExpr: ...
    def __gt__(self, other: Any) -> MolangExpr: ...
    def __lt__(self, other: Any) -> MolangExpr: ...
    def __ge__(self, other: Any) -> MolangExpr: ...
    def __le__(self, other: Any) -> MolangExpr: ...
    def __ne__(self, other: Any) -> MolangExpr: ...  # type: ignore
    def __eq__(self, other: Any) -> MolangExpr: ...  # type: ignore
    def __add__(self, other: Any) -> MolangExpr: ...
    def __sub__(self, other: Any) -> MolangExpr: ...
    def __mul__(self, other: Any) -> MolangExpr: ...
    def __truediv__(self, other: Any) -> MolangExpr: ...
    def __floordiv__(self, other: Any) -> MolangExpr: ...
    def __getitem__(self, name: float | int | MolangExpr) -> MolangExpr: ...
    def if_(self, true_v: Any, false_v: Any) -> MolangExpr: ...
    @staticmethod
    def for_each(
        variable: MolangExpr, array: MolangExpr, expression: MolangExpr
    ) -> MolangExpr: ...
    @staticmethod
    def loop(count: int | MolangExpr, expression: MolangExpr) -> MolangExpr: ...

class VariableNamespace:
    def __call__(self, name: str, *args) -> MolangExpr: ...
    def __init__(self, prefix: str) -> None: ...
    def __getattr__(self, name: str) -> MolangExpr: ...
    def __setattr__(self, name: str, value: Any) -> None: ...

class ArrayNamespace:
    def __getattr__(self, name: str) -> MolangExpr: ...

class MathNamespace(VariableNamespace):
    def abs(self, value) -> MolangExpr: ...
    def acos(self, value) -> MolangExpr: ...
    def asin(self, value) -> MolangExpr: ...
    def atan(self, value) -> MolangExpr: ...
    def atan2(self, y, x) -> MolangExpr: ...
    def ceil(self, value) -> MolangExpr: ...
    def clamp(self, value, min, max) -> MolangExpr: ...
    def copy_sign(self, A, B) -> MolangExpr: ...
    def cos(self, value) -> MolangExpr: ...
    def die_roll(self, num, low, high) -> MolangExpr: ...
    def die_roll_integer(self, num, low, high) -> MolangExpr: ...
    def ease_in_back(self, start, end, value) -> MolangExpr: ...
    def ease_in_bounce(self, start, end, value) -> MolangExpr: ...
    def ease_in_circ(self, start, end, value) -> MolangExpr: ...
    def ease_in_cubic(self, start, end, value) -> MolangExpr: ...
    def ease_in_elastic(self, start, end, value) -> MolangExpr: ...
    def ease_in_expo(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_back(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_bounce(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_circ(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_cubic(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_elastic(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_expo(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_quad(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_quart(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_quint(self, start, end, value) -> MolangExpr: ...
    def ease_in_out_sine(self, start, end, value) -> MolangExpr: ...
    def ease_in_quad(self, start, end, value) -> MolangExpr: ...
    def ease_in_quart(self, start, end, value) -> MolangExpr: ...
    def ease_in_quint(self, start, end, value) -> MolangExpr: ...
    def ease_in_sine(self, start, end, value) -> MolangExpr: ...
    def ease_out_back(self, start, end, value) -> MolangExpr: ...
    def ease_out_bounce(self, start, end, value) -> MolangExpr: ...
    def ease_out_circ(self, start, end, value) -> MolangExpr: ...
    def ease_out_cubic(self, start, end, value) -> MolangExpr: ...
    def ease_out_elastic(self, start, end, value) -> MolangExpr: ...
    def ease_out_expo(self, start, end, value) -> MolangExpr: ...
    def ease_out_quad(self, start, end, value) -> MolangExpr: ...
    def ease_out_quart(self, start, end, value) -> MolangExpr: ...
    def ease_out_quint(self, start, end, value) -> MolangExpr: ...
    def ease_out_sine(self, start, end, value) -> MolangExpr: ...
    def exp(self, value) -> MolangExpr: ...
    def floor(self, value) -> MolangExpr: ...
    def hermite_blend(self, value) -> MolangExpr: ...
    def inverse_lerp(self, start, end, value) -> MolangExpr: ...
    def lerp(self, start, end, value) -> MolangExpr: ...
    def lerprotate(self, start, end, value) -> MolangExpr: ...
    def ln(self, value) -> MolangExpr: ...
    def max(self, A, B) -> MolangExpr: ...
    def min(self, A, B) -> MolangExpr: ...
    def min_angle(self, value) -> MolangExpr: ...
    def mod(self, value, denominator) -> MolangExpr: ...
    def pow(self, base, exponent) -> MolangExpr: ...
    def random(self, low, high) -> MolangExpr: ...
    def random_integer(self, low, high) -> MolangExpr: ...
    def round(self, value) -> MolangExpr: ...
    def sign(self, value) -> MolangExpr: ...
    def sin(self, value) -> MolangExpr: ...
    def sqrt(self, value) -> MolangExpr: ...
    def trunc(self, value) -> MolangExpr: ...

    pi: MolangExpr

class QueryNamespace(VariableNamespace):
    def block_state(self, name: str) -> MolangExpr: ...
    def equipped_item_all_tags(self, slot: str, *tag: str) -> MolangExpr: ...
    def equipped_item_any_tag(self, slot: str, *tag: str) -> MolangExpr: ...
    def get_nearby_entities(self, count: int, type_id: str) -> MolangExpr: ...

    above_top_solid: MolangExpr
    actor_count: MolangExpr
    all: MolangExpr
    all_animations_finished: MolangExpr
    all_tags: MolangExpr
    anger_level: MolangExpr
    anim_time: MolangExpr
    any: MolangExpr
    any_animation_finished: MolangExpr
    any_tag: MolangExpr
    approx_eq: MolangExpr
    armor_color_slot: MolangExpr
    armor_damage_slot: MolangExpr
    armor_material_slot: MolangExpr
    armor_texture_slot: MolangExpr
    average_frame_time: MolangExpr
    base_swing_duration: MolangExpr
    block_face: MolangExpr
    block_has_all_tags: MolangExpr
    block_has_any_tag: MolangExpr
    block_neighbor_has_all_tags: MolangExpr
    block_neighbor_has_any_tag: MolangExpr
    block_property: MolangExpr
    blocking: MolangExpr
    body_x_rotation: MolangExpr
    body_y_rotation: MolangExpr
    bone_aabb: MolangExpr
    bone_orientation_matrix: MolangExpr
    bone_orientation_trs: MolangExpr
    bone_origin: MolangExpr
    bone_rotation: MolangExpr
    camera_distance_range_lerp: MolangExpr
    camera_rotation: MolangExpr
    can_climb: MolangExpr
    can_damage_nearby_mobs: MolangExpr
    can_dash: MolangExpr
    can_fly: MolangExpr
    can_power_jump: MolangExpr
    can_swim: MolangExpr
    can_walk: MolangExpr
    cape_flap_amount: MolangExpr
    cardinal_block_face_placed_on: MolangExpr
    cardinal_facing: MolangExpr
    cardinal_facing_2d: MolangExpr
    cardinal_player_facing: MolangExpr
    client_max_render_distance: MolangExpr
    client_memory_tier: MolangExpr
    combine_entities: MolangExpr
    cooldown_time: MolangExpr
    cooldown_time_remaining: MolangExpr
    count: MolangExpr
    current_squish_value: MolangExpr
    dash_cooldown_progress: MolangExpr
    day: MolangExpr
    death_ticks: MolangExpr
    debug_output: MolangExpr
    delta_time: MolangExpr
    distance_from_camera: MolangExpr
    effect_emitter_count: MolangExpr
    effect_particle_count: MolangExpr
    entity_biome_has_all_tags: MolangExpr
    entity_biome_has_any_identifier: MolangExpr
    entity_biome_has_any_tags: MolangExpr
    equipment_count: MolangExpr
    equipped_item_is_attachable: MolangExpr
    eye_target_x_rotation: MolangExpr
    eye_target_y_rotation: MolangExpr
    facing_target_to_range_attack: MolangExpr
    frame_alpha: MolangExpr
    get_actor_info_id: MolangExpr
    get_animation_frame: MolangExpr
    get_default_bone_pivot: MolangExpr
    get_equipped_item_name: MolangExpr
    get_locator_offset: MolangExpr
    get_name: MolangExpr
    get_pack_setting: MolangExpr
    get_root_locator_offset: MolangExpr
    graphics_mode_is_any: MolangExpr
    ground_speed: MolangExpr
    had_component_group: MolangExpr
    has_any_family: MolangExpr
    has_any_leashed_entity_of_type: MolangExpr
    has_armor_slot: MolangExpr
    has_biome_tag: MolangExpr
    has_block_property: MolangExpr
    has_block_state: MolangExpr
    has_cape: MolangExpr
    has_collision: MolangExpr
    has_dash_cooldown: MolangExpr
    has_gravity: MolangExpr
    has_head_gear: MolangExpr
    has_owner: MolangExpr
    has_player_rider: MolangExpr
    has_property: MolangExpr
    has_rider: MolangExpr
    has_target: MolangExpr
    head_roll_angle: MolangExpr
    head_x_rotation: MolangExpr
    head_y_rotation: MolangExpr
    health: MolangExpr
    heartbeat_interval: MolangExpr
    heartbeat_phase: MolangExpr
    heightmap: MolangExpr
    hurt_direction: MolangExpr
    hurt_time: MolangExpr
    in_range: MolangExpr
    invulnerable_ticks: MolangExpr
    is_admiring: MolangExpr
    is_alive: MolangExpr
    is_angry: MolangExpr
    is_attached: MolangExpr
    is_attached_to_entity: MolangExpr
    is_avoiding_block: MolangExpr
    is_avoiding_mobs: MolangExpr
    is_baby: MolangExpr
    is_breathing: MolangExpr
    is_bribed: MolangExpr
    is_carrying_block: MolangExpr
    is_casting: MolangExpr
    is_celebrating: MolangExpr
    is_celebrating_special: MolangExpr
    is_charged: MolangExpr
    is_charging: MolangExpr
    is_chested: MolangExpr
    is_cooldown_category: MolangExpr
    is_crawling: MolangExpr
    is_critical: MolangExpr
    is_croaking: MolangExpr
    is_dancing: MolangExpr
    is_delayed_attacking: MolangExpr
    is_digging: MolangExpr
    is_eating: MolangExpr
    is_eating_mob: MolangExpr
    is_elder: MolangExpr
    is_emerging: MolangExpr
    is_emoting: MolangExpr
    is_enchanted: MolangExpr
    is_feeling_happy: MolangExpr
    is_fire_immune: MolangExpr
    is_first_person: MolangExpr
    is_ghost: MolangExpr
    is_gliding: MolangExpr
    is_grazing: MolangExpr
    is_idling: MolangExpr
    is_ignited: MolangExpr
    is_illager_captain: MolangExpr
    is_in_contact_with_water: MolangExpr
    is_in_lava: MolangExpr
    is_in_love: MolangExpr
    is_in_ui: MolangExpr
    is_in_water: MolangExpr
    is_in_water_or_rain: MolangExpr
    is_interested: MolangExpr
    is_invisible: MolangExpr
    is_item_equipped: MolangExpr
    is_item_name_any: MolangExpr
    is_jump_goal_jumping: MolangExpr
    is_jumping: MolangExpr
    is_laying_down: MolangExpr
    is_laying_egg: MolangExpr
    is_leashed: MolangExpr
    is_levitating: MolangExpr
    is_lingering: MolangExpr
    is_local_player: MolangExpr
    is_moving: MolangExpr
    is_name_any: MolangExpr
    is_on_fire: MolangExpr
    is_on_ground: MolangExpr
    is_on_screen: MolangExpr
    is_onfire: MolangExpr
    is_orphaned: MolangExpr
    is_owner_identifier_any: MolangExpr
    is_pack_setting_enabled: MolangExpr
    is_pack_setting_selected: MolangExpr
    is_persona_or_premium_skin: MolangExpr
    is_playing_dead: MolangExpr
    is_powered: MolangExpr
    is_pregnant: MolangExpr
    is_ram_attacking: MolangExpr
    is_resting: MolangExpr
    is_riding: MolangExpr
    is_riding_any_entity_of_type: MolangExpr
    is_rising: MolangExpr
    is_roaring: MolangExpr
    is_rolling: MolangExpr
    is_saddled: MolangExpr
    is_scared: MolangExpr
    is_scenting: MolangExpr
    is_searching: MolangExpr
    is_selected_item: MolangExpr
    is_shaking: MolangExpr
    is_shaking_wetness: MolangExpr
    is_sheared: MolangExpr
    is_shield_powered: MolangExpr
    is_silent: MolangExpr
    is_sitting: MolangExpr
    is_sleeping: MolangExpr
    is_sneaking: MolangExpr
    is_sneezing: MolangExpr
    is_sniffing: MolangExpr
    is_sonic_boom: MolangExpr
    is_spectator: MolangExpr
    is_sprinting: MolangExpr
    is_stackable: MolangExpr
    is_stalking: MolangExpr
    is_standing: MolangExpr
    is_stunned: MolangExpr
    is_swimming: MolangExpr
    is_tamed: MolangExpr
    is_transforming: MolangExpr
    is_using_item: MolangExpr
    is_wall_climbing: MolangExpr
    item_in_use_duration: MolangExpr
    item_is_charged: MolangExpr
    item_max_use_duration: MolangExpr
    item_remaining_use_duration: MolangExpr
    item_slot_to_bone_name: MolangExpr
    key_frame_lerp_time: MolangExpr
    kinetic_weapon_damage_duration: MolangExpr
    kinetic_weapon_delay: MolangExpr
    kinetic_weapon_dismount_duration: MolangExpr
    kinetic_weapon_knockback_duration: MolangExpr
    last_frame_time: MolangExpr
    last_hit_by_player: MolangExpr
    last_input_mode_is_any: MolangExpr
    leashed_entity_count: MolangExpr
    lie_amount: MolangExpr
    life_span: MolangExpr
    life_time: MolangExpr
    lod_index: MolangExpr
    log: MolangExpr
    main_hand_item_max_duration: MolangExpr
    main_hand_item_use_duration: MolangExpr
    mark_variant: MolangExpr
    max_durability: MolangExpr
    max_health: MolangExpr
    max_trade_tier: MolangExpr
    maximum_frame_time: MolangExpr
    minimum_frame_time: MolangExpr
    model_scale: MolangExpr
    modified_distance_moved: MolangExpr
    modified_move_speed: MolangExpr
    modified_swing_duration: MolangExpr
    moon_brightness: MolangExpr
    moon_phase: MolangExpr
    movement_direction: MolangExpr
    noise: MolangExpr
    on_fire_time: MolangExpr
    out_of_control: MolangExpr
    overlay_alpha: MolangExpr
    owner_identifier: MolangExpr
    player_level: MolangExpr
    position: MolangExpr
    position_delta: MolangExpr
    previous_squish_value: MolangExpr
    relative_block_has_all_tags: MolangExpr
    relative_block_has_any_tag: MolangExpr
    remaining_durability: MolangExpr
    ride_body_x_rotation: MolangExpr
    ride_body_y_rotation: MolangExpr
    ride_head_x_rotation: MolangExpr
    ride_head_y_rotation: MolangExpr
    rider_body_x_rotation: MolangExpr
    rider_body_y_rotation: MolangExpr
    rider_head_x_rotation: MolangExpr
    rider_head_y_rotation: MolangExpr
    roll_counter: MolangExpr
    rotation_to_camera: MolangExpr
    scoreboard: MolangExpr
    server_memory_tier: MolangExpr
    shake_angle: MolangExpr
    shake_time: MolangExpr
    shield_blocking_bob: MolangExpr
    show_bottom: MolangExpr
    sit_amount: MolangExpr
    skin_id: MolangExpr
    sleep_rotation: MolangExpr
    sneeze_counter: MolangExpr
    spellcolor: MolangExpr
    standing_scale: MolangExpr
    state_time: MolangExpr
    structural_integrity: MolangExpr
    surface_particle_color: MolangExpr
    surface_particle_texture_coordinate: MolangExpr
    surface_particle_texture_size: MolangExpr
    swell_amount: MolangExpr
    swelling_dir: MolangExpr
    swim_amount: MolangExpr
    tail_angle: MolangExpr
    target_x_rotation: MolangExpr
    target_y_rotation: MolangExpr
    texture_frame_index: MolangExpr
    ticks_since_last_kinetic_weapon_hit: MolangExpr
    time_of_day: MolangExpr
    time_since_last_vibration_detection: MolangExpr
    time_stamp: MolangExpr
    timer_flag_1: MolangExpr
    timer_flag_2: MolangExpr
    timer_flag_3: MolangExpr
    total_emitter_count: MolangExpr
    total_particle_count: MolangExpr
    touch_only_affects_hotbar: MolangExpr
    trade_tier: MolangExpr
    unhappy_counter: MolangExpr
    variant: MolangExpr
    vertical_speed: MolangExpr
    walk_distance: MolangExpr
    wing_flap_position: MolangExpr
    wing_flap_speed: MolangExpr
    yaw_speed: MolangExpr

variable: VariableNamespace
temp: VariableNamespace
context: VariableNamespace
geometry: VariableNamespace
texture: VariableNamespace
material: VariableNamespace
query: QueryNamespace
array: ArrayNamespace
math: MathNamespace
